@@ -10,11 +10,10 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use opendal::services::s3;
+use opendal::layers::LoggingLayer;
+use opendal::services::S3;
 use opendal::Operator;
 
-use std::convert::TryInto;
-
 use crate::errors::*;
 
 pub struct S3Cache;
@@ -28,20 +27,34 @@ impl S3Cache {
         endpoint: Option<&str>,
         use_ssl: Option<bool>,
     ) -> Result<Operator> {
-        let mut builder = s3::Builder::default();
+        let mut builder = S3::default();
         builder.bucket(bucket);
+        builder.root(key_prefix);
+
         if let Some(region) = region {
             builder.region(region);
         }
-        builder.root(key_prefix);
+
         if no_credentials {
-            builder.disable_credential_loader();
+            builder.disable_config_load();
+            // Disable EC2 metadata to avoid OpenDAL trying to load
+            // credentials from EC2 metadata.
+            //
+            // A.k.a, don't try to visit `http://169.254.169.254`
+            builder.disable_ec2_metadata();
+            // Allow anonymous access to S3 so that OpenDAL will not
+            // throw error when no credentials are provided.
+            builder.allow_anonymous();
         }
+
         if let Some(endpoint) = endpoint {
             builder.endpoint(&endpoint_resolver(endpoint, use_ssl)?);
         }
 
-        Ok(builder.build()?.into())
+        let op = Operator::new(builder)?
+            .layer(LoggingLayer::default())
+            .finish();
+        Ok(op)
     }
 }
 
