@@ -13,16 +13,20 @@
 // limitations under the License.
 
 use directories::ProjectDirs;
+use fs::File;
+use fs_err as fs;
+use once_cell::sync::Lazy;
 use regex::Regex;
-use serde::de::{Deserialize, DeserializeOwned, Deserializer};
 #[cfg(any(feature = "dist-client", feature = "dist-server"))]
-#[cfg(any(feature = "dist-client", feature = "dist-server"))]
-use serde::ser::{Serialize, Serializer};
+use serde::ser::Serializer;
+use serde::{
+    de::{DeserializeOwned, Deserializer},
+    Deserialize, Serialize,
+};
 #[cfg(test)]
 use serial_test::serial;
 use std::collections::HashMap;
 use std::env;
-use std::fs::{self, File};
 use std::io::{Read, Write};
 use std::path::{Path, PathBuf};
 use std::result::Result as StdResult;
@@ -31,10 +35,8 @@ use std::sync::Mutex;
 
 use crate::errors::*;
 
-lazy_static! {
-    static ref CACHED_CONFIG_PATH: PathBuf = CachedConfig::file_config_path();
-    static ref CACHED_CONFIG: Mutex<Option<CachedFileConfig>> = Mutex::new(None);
-}
+static CACHED_CONFIG_PATH: Lazy<PathBuf> = Lazy::new(CachedConfig::file_config_path);
+static CACHED_CONFIG: Lazy<Mutex<Option<CachedFileConfig>>> = Lazy::new(|| Mutex::new(None));
 
 const ORGANIZATION: &str = "Mozilla";
 const APP_NAME: &str = "sccache";
@@ -192,16 +194,30 @@ pub struct GCSCacheConfig {
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct GHACacheConfig {
-    pub url: String,
-    pub token: String,
-    pub cache_to: Option<String>,
-    pub cache_from: Option<String>,
+    pub enabled: bool,
+    /// Version for gha cache is a namespace. By setting different versions,
+    /// we can avoid mixed caches.
+    pub version: String,
 }
 
+/// Memcached's default value of expiration is 10800s (3 hours), which is too
+/// short for use case of sccache.
+///
+/// We increase the default expiration to 86400s (1 day) to balance between
+/// memory consumpation and cache hit rate.
+///
+/// Please change this value freely if we have a better choice.
+const DEFAULT_MEMCACHED_CACHE_EXPIRATION: u32 = 86400;
+
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct MemcachedCacheConfig {
     pub url: String,
+    /// the expiration time in seconds.
+    ///
+    /// Default to 24 hours (86400)
+    /// Up to 30 days (2592000)
+    pub expiration: u32,
 }
 
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
@@ -210,6 +226,16 @@ pub struct RedisCacheConfig {
     pub url: String,
 }
 
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+#[serde(deny_unknown_fields)]
+pub struct WebdavCacheConfig {
+    pub endpoint: String,
+    pub key_prefix: String,
+    pub username: Option<String>,
+    pub password: Option<String>,
+    pub token: Option<String>,
+}
+
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct S3CacheConfig {
@@ -229,6 +255,7 @@ pub enum CacheType {
     Memcached(MemcachedCacheConfig),
     Redis(RedisCacheConfig),
     S3(S3CacheConfig),
+    Webdav(WebdavCacheConfig),
 }
 
 #[derive(Debug, Default, Serialize, Deserialize, PartialEq, Eq)]
@@ -241,6 +268,7 @@ pub struct CacheConfigs {
     pub memcached: Option<MemcachedCacheConfig>,
     pub redis: Option<RedisCacheConfig>,
     pub s3: Option<S3CacheConfig>,
+    pub webdav: Option<WebdavCacheConfig>,
 }
 
 impl CacheConfigs {
@@ -255,6 +283,7 @@ impl CacheConfigs {
             memcached,
             redis,
             s3,
+            webdav,
         } = self;
 
         let cache_type = s3
@@ -263,7 +292,8 @@ impl CacheConfigs {
             .or_else(|| memcached.map(CacheType::Memcached))
             .or_else(|| gcs.map(CacheType::GCS))
             .or_else(|| gha.map(CacheType::GHA))
-            .or_else(|| azure.map(CacheType::Azure));
+            .or_else(|| azure.map(CacheType::Azure))
+            .or_else(|| webdav.map(CacheType::Webdav));
 
         let fallback = disk.unwrap_or_default();
 
@@ -280,6 +310,7 @@ impl CacheConfigs {
             memcached,
             redis,
             s3,
+            webdav,
         } = other;
 
         if azure.is_some() {
@@ -303,6 +334,9 @@ impl CacheConfigs {
         if s3.is_some() {
             self.s3 = s3
         }
+        if webdav.is_some() {
+            self.webdav = webdav
+        }
     }
 }
 
@@ -474,9 +508,14 @@ pub struct EnvConfig {
 
 fn config_from_env() -> Result<EnvConfig> {
     // ======= AWS =======
-    let s3 = env::var("SCCACHE_BUCKET").ok().map(|bucket| {
+    let s3 = if let Ok(bucket) = env::var("SCCACHE_BUCKET") {
         let region = env::var("SCCACHE_REGION").ok();
-        let no_credentials = env::var("SCCACHE_S3_NO_CREDENTIALS").ok().is_some();
+        let no_credentials =
+            env::var("SCCACHE_S3_NO_CREDENTIALS").map_or(Ok(false), |val| match val.as_str() {
+                "true" | "1" => Ok(true),
+                "false" | "0" => Ok(false),
+                _ => bail!("SCCACHE_S3_NO_CREDENTIALS must be 'true', '1', 'false', or '0'."),
+            })?;
         let use_ssl = env::var("SCCACHE_S3_USE_SSL")
             .ok()
             .map(|value| value != "off");
@@ -489,15 +528,18 @@ fn config_from_env() -> Result<EnvConfig> {
             .map(|s| s.to_owned() + "/")
             .unwrap_or_default();
 
-        S3CacheConfig {
+        Some(S3CacheConfig {
             bucket,
             region,
             no_credentials,
             key_prefix,
             endpoint,
             use_ssl,
-        }
-    });
+        })
+    } else {
+        None
+    };
+
     if s3.as_ref().map(|s3| s3.no_credentials).unwrap_or_default()
         && (env::var_os("AWS_ACCESS_KEY_ID").is_some()
             || env::var_os("AWS_SECRET_ACCESS_KEY").is_some())
@@ -511,9 +553,16 @@ fn config_from_env() -> Result<EnvConfig> {
         .map(|url| RedisCacheConfig { url });
 
     // ======= memcached =======
+    let expiration = match env::var("SCCACHE_MEMCACHED_EXPIRATION").ok() {
+        None => DEFAULT_MEMCACHED_CACHE_EXPIRATION,
+        Some(v) => v
+            .parse()
+            .map_err(|err| anyhow!("SCCACHE_MEMCACHED_EXPIRATION value is invalid: {err:?}"))?,
+    };
+
     let memcached = env::var("SCCACHE_MEMCACHED")
         .ok()
-        .map(|url| MemcachedCacheConfig { url });
+        .map(|url| MemcachedCacheConfig { url, expiration });
 
     // ======= GCP/GCS =======
     if (env::var("SCCACHE_GCS_CREDENTIALS_URL").is_ok()
@@ -538,8 +587,8 @@ fn config_from_env() -> Result<EnvConfig> {
 
 
         if env::var("SCCACHE_GCS_OAUTH_URL").is_ok() {
-            warn!("SCCACHE_GCS_OAUTH_URL has been deprecated");
-            warn!("if you intend to use vm metadata for auth, please set correct service account intead");
+            eprintln!("SCCACHE_GCS_OAUTH_URL has been deprecated");
+            eprintln!("if you intend to use vm metadata for auth, please set correct service account intead");
         }
 
         let credential_url = env::var("SCCACHE_GCS_CREDENTIALS_URL").ok();
@@ -573,20 +622,24 @@ fn config_from_env() -> Result<EnvConfig> {
     });
 
     // ======= GHA =======
-    let gha = if let (Some(url), Some(token)) = (
-        env::var("SCCACHE_GHA_CACHE_URL")
-            .ok()
-            .or_else(|| env::var("ACTIONS_CACHE_URL").ok()),
-        env::var("SCCACHE_GHA_RUNTIME_TOKEN")
-            .ok()
-            .or_else(|| env::var("ACTIONS_RUNTIME_TOKEN").ok()),
-    ) {
+    let gha = if let Ok(version) = env::var("SCCACHE_GHA_VERSION") {
+        // If SCCACHE_GHA_VERSION has been set, we don't need to check
+        // SCCACHE_GHA_ENABLED's value anymore.
         Some(GHACacheConfig {
-            url,
-            token,
-            cache_to: env::var("SCCACHE_GHA_CACHE_TO").ok(),
-            cache_from: env::var("SCCACHE_GHA_CACHE_FROM").ok(),
+            enabled: true,
+            version,
         })
+    } else if let Ok(enabled) = env::var("SCCACHE_GHA_ENABLED") {
+        // If only SCCACHE_GHA_ENABLED has been set, enable with
+        // default version.
+        if enabled == "on" || enabled == "true" {
+            Some(GHACacheConfig {
+                enabled: true,
+                version: "".to_string(),
+            })
+        } else {
+            None
+        }
     } else {
         None
     };
@@ -612,6 +665,30 @@ fn config_from_env() -> Result<EnvConfig> {
         None
     };
 
+    // ======= WebDAV =======
+    let webdav = if let Ok(endpoint) = env::var("SCCACHE_WEBDAV_ENDPOINT") {
+        let key_prefix = env::var("SCCACHE_WEBDAV_KEY_PREFIX")
+            .ok()
+            .as_ref()
+            .map(|s| s.trim_end_matches('/'))
+            .filter(|s| !s.is_empty())
+            .unwrap_or_default()
+            .to_owned();
+        let username = env::var("SCCACHE_WEBDAV_USERNAME").ok();
+        let password = env::var("SCCACHE_WEBDAV_PASSWORD").ok();
+        let token = env::var("SCCACHE_WEBDAV_TOKEN").ok();
+
+        Some(WebdavCacheConfig {
+            endpoint,
+            key_prefix,
+            username,
+            password,
+            token,
+        })
+    } else {
+        None
+    };
+
     // ======= Local =======
     let disk_dir = env::var_os("SCCACHE_DIR").map(PathBuf::from);
     let disk_sz = env::var("SCCACHE_CACHE_SIZE")
@@ -635,6 +712,7 @@ fn config_from_env() -> Result<EnvConfig> {
         memcached,
         redis,
         s3,
+        webdav,
     };
 
     Ok(EnvConfig { cache })
@@ -788,7 +866,7 @@ impl CachedConfig {
     fn save_file_config(c: &CachedFileConfig) -> Result<()> {
         let file_conf_path = &*CACHED_CONFIG_PATH;
         let mut file = File::create(file_conf_path).context("Could not open config for writing")?;
-        file.write_all(&toml::to_vec(c).unwrap())
+        file.write_all(toml::to_string(c).unwrap().as_bytes())
             .map_err(Into::into)
     }
 }
@@ -800,6 +878,8 @@ pub mod scheduler {
 
     use crate::errors::*;
 
+    use serde::{Deserialize, Serialize};
+
     #[derive(Debug, Serialize, Deserialize)]
     #[serde(tag = "type")]
     #[serde(deny_unknown_fields)]
@@ -851,6 +931,7 @@ pub mod scheduler {
 #[cfg(feature = "dist-server")]
 pub mod server {
     use super::HTTPUrl;
+    use serde::{Deserialize, Serialize};
     use std::net::SocketAddr;
     use std::path::{Path, PathBuf};
 
@@ -976,6 +1057,7 @@ fn config_overrides() {
             }),
             memcached: Some(MemcachedCacheConfig {
                 url: "memurl".to_owned(),
+                expiration: 24 * 3600,
             }),
             redis: Some(RedisCacheConfig {
                 url: "myredisurl".to_owned(),
@@ -1004,8 +1086,8 @@ fn config_overrides() {
 
 #[test]
 #[serial]
-fn test_s3_no_credentials() {
-    env::set_var("SCCACHE_S3_NO_CREDENTIALS", "1");
+fn test_s3_no_credentials_conflict() {
+    env::set_var("SCCACHE_S3_NO_CREDENTIALS", "true");
     env::set_var("SCCACHE_BUCKET", "my-bucket");
     env::set_var("AWS_ACCESS_KEY_ID", "aws-access-key-id");
     env::set_var("AWS_SECRET_ACCESS_KEY", "aws-secret-access-key");
@@ -1022,6 +1104,68 @@ fn test_s3_no_credentials() {
     env::remove_var("AWS_SECRET_ACCESS_KEY");
 }
 
+#[test]
+#[serial]
+fn test_s3_no_credentials_invalid() {
+    env::set_var("SCCACHE_S3_NO_CREDENTIALS", "yes");
+    env::set_var("SCCACHE_BUCKET", "my-bucket");
+
+    let error = config_from_env().unwrap_err();
+    assert_eq!(
+        "SCCACHE_S3_NO_CREDENTIALS must be 'true', '1', 'false', or '0'.",
+        error.to_string()
+    );
+
+    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
+    env::remove_var("SCCACHE_BUCKET");
+}
+
+#[test]
+#[serial]
+fn test_s3_no_credentials_valid_true() {
+    env::set_var("SCCACHE_S3_NO_CREDENTIALS", "true");
+    env::set_var("SCCACHE_BUCKET", "my-bucket");
+
+    let env_cfg = config_from_env().unwrap();
+    match env_cfg.cache.s3 {
+        Some(S3CacheConfig {
+            ref bucket,
+            no_credentials,
+            ..
+        }) => {
+            assert_eq!(bucket, "my-bucket");
+            assert!(no_credentials);
+        }
+        None => unreachable!(),
+    };
+
+    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
+    env::remove_var("SCCACHE_BUCKET");
+}
+
+#[test]
+#[serial]
+fn test_s3_no_credentials_valid_false() {
+    env::set_var("SCCACHE_S3_NO_CREDENTIALS", "false");
+    env::set_var("SCCACHE_BUCKET", "my-bucket");
+
+    let env_cfg = config_from_env().unwrap();
+    match env_cfg.cache.s3 {
+        Some(S3CacheConfig {
+            ref bucket,
+            no_credentials,
+            ..
+        }) => {
+            assert_eq!(bucket, "my-bucket");
+            assert!(!no_credentials);
+        }
+        None => unreachable!(),
+    };
+
+    env::remove_var("SCCACHE_S3_NO_CREDENTIALS");
+    env::remove_var("SCCACHE_BUCKET");
+}
+
 #[test]
 fn test_gcs_service_account() {
     env::set_var("SCCACHE_GCS_BUCKET", "my-bucket");
@@ -1083,13 +1227,12 @@ key_prefix = "prefix"
 service_account = "example_service_account"
 
 [cache.gha]
-url = "http://localhost"
-token = "secret"
-cache_to = "sccache-latest"
-cache_from = "sccache-"
+enabled = true
+version = "sccache"
 
 [cache.memcached]
 url = "..."
+expiration = 86400
 
 [cache.redis]
 url = "redis://user:passwd@1.2.3.4:6379/1"
@@ -1101,6 +1244,13 @@ endpoint = "s3-us-east-1.amazonaws.com"
 use_ssl = true
 key_prefix = "s3prefix"
 no_credentials = true
+
+[cache.webdav]
+endpoint = "http://127.0.0.1:8080"
+key_prefix = "webdavprefix"
+username = "webdavusername"
+password = "webdavpassword"
+token = "webdavtoken"
 "#;
 
     let file_config: FileConfig = toml::from_str(CONFIG_STR).expect("Is valid toml.");
@@ -1122,16 +1272,15 @@ no_credentials = true
                     credential_url: None,
                 }),
                 gha: Some(GHACacheConfig {
-                    url: "http://localhost".to_owned(),
-                    token: "secret".to_owned(),
-                    cache_to: Some("sccache-latest".to_owned()),
-                    cache_from: Some("sccache-".to_owned()),
+                    enabled: true,
+                    version: "sccache".to_string()
                 }),
                 redis: Some(RedisCacheConfig {
                     url: "redis://user:passwd@1.2.3.4:6379/1".to_owned(),
                 }),
                 memcached: Some(MemcachedCacheConfig {
                     url: "...".to_owned(),
+                    expiration: 24 * 3600
                 }),
                 s3: Some(S3CacheConfig {
                     bucket: "name".to_owned(),
@@ -1141,6 +1290,13 @@ no_credentials = true
                     key_prefix: "s3prefix".into(),
                     no_credentials: true,
                 }),
+                webdav: Some(WebdavCacheConfig {
+                    endpoint: "http://127.0.0.1:8080".to_string(),
+                    key_prefix: "webdavprefix".into(),
+                    username: Some("webdavusername".to_string()),
+                    password: Some("webdavpassword".to_string()),
+                    token: Some("webdavtoken".to_string()),
+                })
             },
             dist: DistConfig {
                 auth: DistAuth::Token {
